/********************************************************************
** Likelihood Evaluation using Kalman Filter
**
** {loglh,retcode,obserror,obsvar} = evalmod(para,YY,msel);
** {loglh,retcode,obserror,obsvar} = evalmod3(para,YY,msel);
**
** #include ****_mod.src,  to run this procedure
**
** Last updated: 06/16/2005 
********************************************************************/


proc (4) = evalmod(para,YY,msel);

local loglh,retcode,obserror,obsvar,loglhzero,
      nobs,ny,nstate,nz,nep,tcond,
      TT,QQ,RR,HH,DD,ZZ,VV,RC,
      t,At,Pt,At1,Pt1,Ft,alphahat,Phat,nut,yhat;

/** likelihood conditional on first tcond observations **/
/** to make it comparable to VAR **/
tcond = 4;      

/** initialize **/
nobs = rows(YY);
ny   = cols(YY);

loglhzero = -1E8;

loglh     = 0;
retcode   = 0;
obserror  = zeros(nobs,ny);
obsvar    = zeros(nobs,ny);


/**********************************************************
**      System matrices for State-space
**********************************************************/
{TT,QQ,RR,HH,DD,ZZ,VV,RC}=sysmat(para,msel);

/**********************************************************
** Check determinacy 
**********************************************************/
if (RC[1] == 1) AND (RC[2]==1);
   /* determinacy */
   retcode[1] = 0;

elseif (RC[1] == 1) AND (RC[2]==0) ;
   /* indeterminacy */
   retcode[1] = 1;
   loglh = loglhzero;
   goto eofeval;

else;
   /* no equilibrium exists, numerical problems */
   retcode[1] = RC[1];
   loglh = loglhzero;
   goto eofeval;

endif;


/**********************************************************
**      Likelihood evaluation by Kalman filter
**********************************************************/
nstate = rows(TT);

/** initial mean and variance for the state vector **/
At = zeros(nstate,1);
Pt = dlyap(TT,RR*QQ*RR');

/** filter loop **/
t = 1;
do until t > nobs;

   At1 = At;
   Pt1 = Pt;

   /** Forecasting **/
   alphahat = TT*At1;
   Phat     = TT*Pt1*TT' + RR*QQ*RR';
   Phat     = .5*(Phat+Phat');

   yhat     = ZZ*alphahat + DD;
   nut      = YY[t,.] - yhat';
   Ft       = ZZ*Phat*ZZ' + HH + ZZ*RR*VV + (ZZ*RR*VV)';
   Ft       = .5*(Ft+Ft');
   
   /** to make likelihood comparable to VAR(tcond), condition on first tcond observations **/
   if t > tcond;
    loglh    =  loglh -0.5*cols(YY)*ln(2*pi) - 0.5*ln(det(Ft)) - 0.5*nut*inv(Ft)*nut';
   endif;

   /** Updating **/
   At = alphahat + Phat*ZZ'*inv(Ft)*nut';
   Pt = Phat - (Phat*ZZ'+RR*VV)*inv(Ft)*(Phat*ZZ'+RR*VV)';

   obserror[t,.] = nut;
   obsvar[t,.]   = diag(Ft)';

   t = t+1;
endo;


eofeval:

if ismiss(real(loglh))==1;
   loglh = loglhzero;
endif;

retp(real(loglh),retcode,obserror,obsvar);
endp;







/********************************************************************/
proc (4) = evalmod2(para,YY,msel);

local loglh,retcode,obserror,obsvar,loglhzero,
      nobs,ny,nstate,nz,nep,tcond,
      TT,QQ,RR,HH,DD,ZZ,VV,RC,
      t,At,Pt,Ptchol,epst,npart,loglh_max,loglh_dev,
      Tqz, Sqz, Qqz, Zqz, lambdaqz, RRqz, nstate_rr, wtoalp, alptow,i;

/** likelihood conditional on first tcond observations **/
/** to make it comparable to VAR **/
tcond = 0;      

/** initialize **/
nobs = rows(YY);
ny   = cols(YY);

loglhzero = -1E8;

loglh     = 0;
retcode   = 0;
obserror  = zeros(nobs,ny);
obsvar    = zeros(nobs,ny);


/**********************************************************
**      System matrices for State-space
**********************************************************/
{TT,QQ,RR,HH,DD,ZZ,VV,RC}=sysmat(para,msel);

/**********************************************************
** Check determinacy 
**********************************************************/
if (RC[1] == 1) AND (RC[2]==1);
   /* determinacy */
   retcode[1] = 0;

elseif (RC[1] == 1) AND (RC[2]==0) ;
   /* indeterminacy */
   retcode[1] = 1;
   loglh = loglhzero;
   goto eofeval;

else;
   /* no equilibrium exists, numerical problems */
   retcode[1] = RC[1];
   loglh = loglhzero;
   goto eofeval;

endif;


/**********************************************************
**      Likelihood evaluation 
**********************************************************/
nstate = rows(TT);

/* goto endoftrans; */
/* Transform system */

/* Schur decomposition, zero eigenvalues last */
{ Tqz,Sqz,Qqz,Zqz,lambdaqz } = dTgSen( TT,eye(nstate),1E-15,1,1 );

/* calculate rank */
nstate_rr = sumc( abs(diag(Tqz)) .> 1E-15 );

/* define w(t) = ZZqz'*s(t)
** w(t) = Tqz*w(t-1) + RRqz*eps(t)
*/
RRqz = Zqz'*RR;
ZZ   = ZZ*Zqz;

/* eliminate unnecessary states: w(t) -> alp(t)
*/
wtoalp = eye(nstate_rr);
alptow = eye(nstate_rr);

if nstate_rr < nstate;

   i = nstate_rr + 1;
   do until i > nstate;
      if sumc(RRqz[i,.]^2) < 1E-15;
         /* eliminate state */
         alptow = alptow | zeros(1,cols(alptow));
         wtoalp = wtoalp ~ zeros(rows(wtoalp),1);
      else;
         /* retain state */
         alptow = (alptow ~ zeros(rows(alptow),1)) |
                  (zeros(1, cols(alptow))   ~ 1  ) ;
         wtoalp = (wtoalp ~ zeros(rows(wtoalp),1)) |
                  (zeros(1, cols(wtoalp))   ~ 1  ) ;
      endif;
      i = i+1;
   endo;
   
endif;

/* transform system into alphas
*/
TT=wtoalp*Tqz*alptow;
RR=wtoalp*RRqz;
ZZ=ZZ*alptow;
nstate = rows(TT);

endoftrans:

/** initial mean and variance for the state vector **/
At = zeros(nstate,1);
Pt = dlyap(TT,RR*QQ*RR');

/** initialize particle filter **/
npart = 100000;
rndseed 1234;
loglh = zeros(npart,1);

Ptchol = cholpsd(Pt);
At = Ptchol*rndn(nstate,npart);

/** filter loop **/
t = 1;
do until t > nobs;

   /* calculate the residual */
   epst = inv(ZZ*RR)*(YY[t,.]'-DD-ZZ*TT*At);
   
   /* likelihood contribution */
      
   if t > tcond;
     loglh = loglh - 0.5*ny*ln(2*pi)
                   - 0.5*ln(det(ZZ*RR*QQ*RR'*ZZ'))
                   - 0.5*sumc(((epst'*inv(QQ)).*epst')');
   endif;
              
   /* update the state vector */
   At = TT*At + RR*epst;

   t = t+1;
endo;

loglh_max = maxc(loglh);
loglh_dev = loglh - loglh_max;

loglh = loglh_max + ln(meanc(exp(loglh_dev)));


eofeval:

if ismiss(real(loglh))==1;
   loglh = loglhzero;
endif;

retp(real(loglh),retcode,obserror,obsvar);
endp;


/********************************************************************/
proc (4) = evalmod3(para,YY,msel);

local loglh,retcode,obserror,obsvar,loglhzero,
      nobs,ny,nstate,nz,nep,tcond,
      TT,QQ,RR,HH,DD,ZZ,VV,RC,
      t,At,Pt,Ptchol,epst,npart,
      loglhpart,loglhpart_max,loglhpart_dev,
      lhpart, lhAt_tr, rnduvec, selind,
      Tqz, Sqz, Qqz, Zqz, lambdaqz, RRqz, nstate_rr, wtoalp, alptow,i;

/** likelihood conditional on first tcond observations **/
/** to make it comparable to VAR **/
tcond = 0;      

/** initialize **/
nobs = rows(YY);
ny   = cols(YY);

loglhzero = -1E8;

loglh     = 0;
retcode   = 0;
obserror  = zeros(nobs,ny);
obsvar    = zeros(nobs,ny);


/**********************************************************
**      System matrices for State-space
**********************************************************/
{TT,QQ,RR,HH,DD,ZZ,VV,RC}=sysmat(para,msel);

/**********************************************************
** Check determinacy 
**********************************************************/
if (RC[1] == 1) AND (RC[2]==1);
   /* determinacy */
   retcode[1] = 0;

elseif (RC[1] == 1) AND (RC[2]==0) ;
   /* indeterminacy */
   retcode[1] = 1;
   loglh = loglhzero;
   goto eofeval;

else;
   /* no equilibrium exists, numerical problems */
   retcode[1] = RC[1];
   loglh = loglhzero;
   goto eofeval;

endif;


/**********************************************************
**      Likelihood evaluation
**********************************************************/
nstate = rows(TT);

/* goto endoftrans; */
/* Transform system */

/* Schur decomposition, zero eigenvalues last */
{ Tqz,Sqz,Qqz,Zqz,lambdaqz } = dTgSen( TT,eye(nstate),1E-15,1,1 );

/* calculate rank */
nstate_rr = sumc( abs(diag(Tqz)) .> 1E-15 );

/* define w(t) = ZZqz'*s(t)
** w(t) = Tqz*w(t-1) + RRqz*eps(t)
*/
RRqz = Zqz'*RR;
ZZ   = ZZ*Zqz;

/* eliminate unnecessary states: w(t) -> alp(t)
*/
wtoalp = eye(nstate_rr);
alptow = eye(nstate_rr);

if nstate_rr < nstate;

   i = nstate_rr + 1;
   do until i > nstate;
      if sumc(RRqz[i,.]^2) < 1E-15;
         /* eliminate state */
         alptow = alptow | zeros(1,cols(alptow));
         wtoalp = wtoalp ~ zeros(rows(wtoalp),1);
      else;
         /* retain state */
         alptow = (alptow ~ zeros(rows(alptow),1)) |
                  (zeros(1, cols(alptow))   ~ 1  ) ;
         wtoalp = (wtoalp ~ zeros(rows(wtoalp),1)) |
                  (zeros(1, cols(wtoalp))   ~ 1  ) ;
      endif;
      i = i+1;
   endo;
   
endif;

/* transform system into alphas
*/
TT=wtoalp*Tqz*alptow;
RR=wtoalp*RRqz;
ZZ=ZZ*alptow;
nstate = rows(TT);

endoftrans:

/** initial mean and variance for the state vector **/
At = zeros(nstate,1);
Pt = dlyap(TT,RR*QQ*RR');

/** initialize particle filter **/
npart = 5000;
rndseed 1234;
loglh = 0;

Ptchol = cholpsd(Pt);
At = Ptchol*rndn(nstate,npart);

/** filter loop **/
t = 1;
do until t > nobs;

   /* calculate the residual */
   epst = inv(ZZ*RR)*(YY[t,.]'-DD-ZZ*TT*At);
   
   /* likelihood */
   
   loglhpart = - 0.5*ny*ln(2*pi)
               - 0.5*ln(det(ZZ*RR*QQ*RR'*ZZ'))
               - 0.5*sumc(((epst'*inv(QQ)).*epst')');
      
   if t > tcond;
     
     loglhpart_max = maxc(loglhpart);
     loglhpart_dev = loglhpart - loglhpart_max;
     loglh = loglh + loglhpart_max + ln(meanc(exp(loglhpart_dev)));
     
   endif;
              
   /* update the state vector for each particle */
   At = TT*At + RR*epst;

   /* compute posterior weight for particles
   */
   lhpart    = exp(loglhpart)/sumc(exp(loglhpart));
   lhAt_tr   = lhpart ~ At';
   
   /* eliminate particles with small weight
   */
   lhAt_tr   = delif(lhAt_tr,( lhAt_tr[.,1] .< (1/npart)*1E-6 ));
     
   /* renormalize weights
   */
   lhAt_tr[.,1] = lhAt_tr[.,1]/sumc(lhAt_tr[.,1]);
   
   /* create cumulative sums
   */
   lhAt_tr[.,1] = cumsumc(lhAt_tr[.,1]);
      
   /* resample particles
   */
   rnduvec = rndu(1,npart);
   selind  = rows(lhAt_tr) - sumc( lhAt_tr[.,1] .> rnduvec ) + 1;
   At      = ( lhAt_tr[selind,2:cols(lhAt_tr)] )';
      
   t = t+1;
endo;



eofeval:

if ismiss(real(loglh))==1;
   loglh = loglhzero;
endif;

retp(real(loglh),retcode,obserror,obsvar);
endp;
